# -*- coding: utf-8 -*-

# arteconcert.py
# This file is part of Qarte-3
#    
# Author: Vincent Vande Vyvre <vincent.vandevyvre@oqapy.eu>
# Copyright: 2011-2016 Vincent Vande Vyvre
# Licence: GPL3
# Home page: https://launchpad.net/qarte
#
# arte Concert main class

import os
import sys
import shutil
import time
import pickle
import json
import glob
import urllib.request
from urllib.error import URLError
from html.parser import HTMLParser
import contextlib
import html
import re
import logging
lgg = logging.getLogger(__name__)

from datetime import datetime
from threading import Thread
 
from PyQt5.QtGui import QFont, QTextCharFormat
from PyQt5.QtCore import QObject, QCoreApplication, pyqtSignal, QLocale, QSize

from loader import Loader
from tvdownloadingconfig import TvDownloadingConfig
from gui.warning import WarningBox
from data import USER_AGENT


class ArteConcert(QObject):
    """Define the arte Concert server handler.

    """
    parsingFinished = pyqtSignal()
    concertsCompleted = pyqtSignal()
    def __init__(self, core):
        super().__init__()
        self.core = core
        self.ui = core.ui
        self.viewer = self.ui.viewer
        self.lang = core.cfg['lang']
        self.default_quality = core.cfg['live_quality']
        self.fatal_error = False
        self.cache = core.workspace['live_tmp']
        self.concerts = []
        self.to_download = []
        self.categories = ['recents', 'classic', 'jazz',
                           'rock', 'dance', 'world']
        self.video_items = {'recents': [], 'classic': [], 'jazz': [], 
                                 'rock': [], 'dance': [], 'world': []}
        self.is_loading = False
        self.loading_aborted = False
        lang = QLocale.system().name()
        self.locale = QLocale(lang)
        
        self.concertsCompleted.connect(self.show_recents)
        for button in self.ui.category_buttons:
            button.triggered.connect(self.change_category)

        self.loader = Loader()
        self.loader.loadingProgress.connect(self.display_progress)
        self.loader.loadingFinished.connect(self.on_downloading_finished)

    def config_parser(self):
        """Configure the main pages parser.

        """
        self.update_concerts()
        self.get_old_concerts()
        self.parsingFinished.emit()

    def update_concerts(self):
        fpath = self.core.workspace['concerts_%s' % self.lang]
        cnt = self.get_update()
        if cnt:
            lgg.info('Concert list updated')
            with open(fpath, 'w') as outf:
                outf.write(cnt)

    def get_update(self):
        url = 'http://www.oqapy.eu/releases/concerts_%s' % self.lang
        req = urllib.request.Request(url, data=None, 
                                     headers={"User-Agent": USER_AGENT})
        try:
            content = urllib.request.urlopen(req)
            return str(content.read().decode('utf-8', 'replace'))
        except Exception as why:
            lgg.info('urllib in update concert error: %s, %s' % (url, why))
            return False

    def get_old_concerts(self):
        fpath = self.core.workspace['concerts_%s' % self.lang]
        if os.path.isfile(fpath):
            try:
                with open(fpath, 'r') as objfile:
                    cnt = objfile.read()
                    self.concerts =  json.loads(cnt)
            except Exception as why:
                lgg.warning("Can't read concert_%s: %s" %(self.lang, why))
                self.concerts = {'recents': [], 'classic': [], 'jazz': [], 
                                 'rock': [], 'dance': [], 'world': []}

    def get_site_pages(self):
        if self.lang == 'fr':
            from data import URLS_FR as urls

        else:
            from data import URLS_DE as urls

        links = []
        for key in ['Classique', 'Jazz & Blues', 'Pop rock & Electro', 
                    'Théâtre & Danse', 'Musiques du monde']:
            links.append(url[key][0])
        return links

    def display_available_concerts(self):
        def build_items():
            for key in self.concerts:
                for conc in self.concerts[key]:
                    ci = ConcertItem(key, conc, self.lang)
                    self.video_items[key].append(ci)
                    self.check_thumbnails(ci)
            self.concertsCompleted.emit()
        Thread(target=build_items).start()

    def show_recents(self):
        for button in self.ui.category_buttons:
            button.setEnabled(True)
        self.viewer.show_category(self.video_items['recents'])
        self.ui.set_active_category(0)

    def change_category(self, category):
        self.viewer.show_category(self.video_items[category])
        self.ui.set_active_category(self.categories.index(category))

    def check_thumbnails(self, item):
        if item.preview is not None:
            if os.path.isfile(item.preview):
                return

        ext = os.path.splitext(item.imgurl)[1] or ".jpg"
        dest = os.path.join(self.cache, "%s%s" %(item.order, ext))
        if not os.path.isfile(dest):
            item.preview = self.fetch_thumbnail(dest, item.imgurl)

        else:
            item.preview = dest

    def fetch_thumbnail(self, fname, url):
        for i in range(2):
            try:
                with open(fname, 'wb') as objfile:
                    f = urllib.request.urlopen(url)
                    objfile.write(f.read())
                return fname
            except Exception as why:
                lgg.info("Can't load image: %s" % url)
                lgg.info("Reason: %s" % why)
                # API error code 104 'Connection reset by peer'
                if not 'Errno 104' in why:
                    break

        return "medias/noPreview.png"

    def get_page(self, url):
        """Get a web page.

        Args:
        url -- the url of the page
        """
        lgg.info('Load page: %s' % url)
        try:
            content = urllib.request.urlopen(url).read()
            return str(content.decode('utf-8', 'replace'))
        except Exception as why:
            lgg.info('urllib error: %s, %s' % (url, why))
            self.urlliberror = (url, why)
            return False

    def show_pitch(self, item):
        """Called when the current item in the viewer has changed.

        """
        title = item.title
        date = "%s %s" %(item.format_date(), item.format_duration())
        pitch = item.summary or item.teaser
        self.update_pitch(title, date, pitch)
 
    def update_pitch(self, *args):
        """Show the pitch of the current item.

        """
        self.ui.editor_live.clear()
        font = QFont("sans", 10)
        c_form = QTextCharFormat()
        c_form.setFont(font)
        font1 = QFont("sans", 12)
        font1.setWeight(75)
        c_form1 = QTextCharFormat()
        c_form1.setFont(font1)
        self.ui.editor_live.setCurrentCharFormat(c_form1)
        self.ui.editor_live.appendPlainText(args[0])
        self.ui.editor_live.setCurrentCharFormat(c_form) 
        self.ui.editor_live.appendPlainText(args[1])
        self.ui.editor_live.appendPlainText("")
        self.ui.editor_live.appendPlainText(args[2])
        self.ui.editor_live.verticalScrollBar().setValue(0)

    def on_thumbnail_selected(self, item):
        # Check if video is already in the downloading basket
        #self.ui.select_btn.setEnabled(not item in self.to_download)
        # and if already available
        self.ui.select_btn.setEnabled(True)

    def select_video_to_download(self, idx=None, basket=True):
        """Select a video for the downloading basket.

        Args:
        idx -- index of video if called by a double click on a item
        """
        if idx is None:
            idx = self.viewer.current

        else:
            self.viewer.go_to_item(idx)

        video = self.viewer.videos_items[idx]
        if video in self.to_download:
            return

        lgg.info("Append item {0} to download basket".format(video.order))
        qualities = self.get_stream_urls(video)

        if qualities is None:
            lgg.warning("No stream URLs found, item rejected")
            return

        self.clean_qualities(qualities)
        video.qualities = qualities
        video.set_file_name()
        video.set_quality(self.default_quality)
        if basket:
            self.to_download.append(video)
            self.ui.live_basket.add_concert(video, self.viewer.previews[idx].img)
            self.ui.select_btn.setEnabled(False)
            self.ui.dl_live_wdg.dwnld_btn.setEnabled(True)
            lgg.info("Found qualities: %s" % (", ".join(qualities.keys())))
            self.ui.actionDo_wnload.setEnabled(True)

    def remove_item_from_basket(self, idx=None):
        """Called if user remove an item from downloading basket or when
        a downloading is complete.

        idx -- index of item
        """
        logger.info('Remove item %s frombasket' % idx)
        if idx == 'first':
            idx = 0

        elif idx is None:
            idx = self.get_current_item_index()

        self.ui.live_basket.delete_item(idx)
        self.to_download.pop(idx)

    def get_stream_urls(self, item):
        content = self.get_page(item.jsonurl)
        if not content:
            lgg.info("Can't get the stream urls page: %s" % self.urlliberror)
            warn = WarningBox(5, why, self.ui)
            rep = warn.exec_()
            return

        try:
            dct = json.loads(content)
            return dct["videoJsonPlayer"]['VSR']
        except Exception as why:
            lgg.info("Can't read the stream urls: %s" % why)
            warn = WarningBox(5, why, self.ui)
            rep = warn.exec_()
            return

    def clean_qualities(self, dct):
        keys = list(dct.keys())
        for key in keys:
            if not 'HTTP' in key:
                dct.pop(key)

    def on_empty_list(self):
        self.ui.dl_live_wdg.dwnld_btn.setEnabled(False)
        self.ui.actionDo_wnload.setEnabled(False)
        self.ui.actionCa_ncel.setEnabled(False)

    def configure_downloading(self, video):
        """Call the downloading config dialog box.

        Args:
        video -- ConcertItem instance
        """
        lgg.info('Configure downloading of %s' % video.title)
        if video.qualities is None:
            self.select_video_to_download(None, False)

        self.cfg = TvDownloadingConfig(self)
        self.cfg.configure(video.qualities, video, 
                           self.core.cfg['live_quality'], 
                           self.core.cfg['lang'], sort=False)
        self.cfg.exec_()

    def set_preferred_quality(self, quality):
        self.core.cfg['live_quality'] = quality

    def start_download(self):
        if not self.to_download:
            self.ui.dl_live_wdg.clear(True)
            return

        video = self.to_download.pop(0)
        lgg.info('Start download: %s' % video.title)
        stream = self.get_stream(video)
        target = self.core.get_target_path(self.core.cfg['music_folder'])
        self.video_loaded = self.get_video_path(video)
        if not self.initiate_loader(stream, target):
            return

        if not self.core.check_free_space(self.file_size, 
                                          self.core.cfg["music_folder"]):
            self.reset_loading_widgets()
            self.to_download.insert(0, video)
            return

        self.update_widgets_on_loading()
        lgg.info('Start download %s quality: %s file size: %s'
                    %(video.fname, video.quality, self.file_size))
        self.is_loading = True
        Thread(target=self.loader.download_mp4).start()

    def get_stream(self, item):
        return item.qualities[item.quality]["url"]

    def initiate_loader(self, url, fname):
        self.file_size = self.loader.config_loader(url, fname)
        if isinstance(self.file_size, int):
            return True

        lgg.info('Loading error: %s' % self.file_size)
        return False

    def update_widgets_on_loading(self):
        self.ui.live_basket.disable_settings()
        self.ui.dl_live_wdg.on_progress()
        self.ui.actionDo_wnload.setEnabled(False)
        self.ui.actionCa_ncel.setEnabled(True)

    def reset_loading_widgets(self):
        self.ui.dl_live_wdg.dwnld_btn.setEnabled(True)
        self.ui.dl_live_wdg.cancel_btn.setEnabled(False)
        self.ui.actionDo_wnload.setEnabled(True)
        self.ui.actionCa_ncel.setEnabled(False)

    def get_video_path(self, item):
        """Return the path/name of the video file.

        Args:
        item -- TVItem instance
        """
        fld = self.core.cfg['music_folder']
        path = os.path.join(fld, item.fname + ".mp4")
        count = 1
        while os.path.isfile(path):
            path = os.path.join(fld, item.fname + "(%s).mp4" % count)
            count += 1

        return path

    def cancel_download(self):
        lgg.info('Cancel download')
        self.loading_aborted = True
        self.loader.kill_request = True

    def display_progress(self, percent, speed, remain):
        """Show the downloading progress informations.

        Args:
        percent -- percent downloaded
        speed -- bandwidth in Ko/sec.
        remain -- estimated remaining time
        """
        rtime = int(remain / speed)
        speed = self.locale.toString(int(speed / 1024))
        self.ui.dl_live_wdg.progressBar.setValue(percent)
        self.ui.dl_live_wdg.progress_lbl.setText(
                                "Speed: %s Kio/sec." % speed)
        m, s = divmod(rtime, 60)
        self.ui.dl_live_wdg.progress_lbl2.setText(
                                "Remaining time: %s min. %s sec." %(m, s))

    def on_downloading_finished(self, fname):
        """Rename the video downloaded.

        Args:
        fname -- temporary file name
        """
        lgg.info('File %s downloaded' % fname)
        self.is_loading = False
        item = self.ui.live_basket.remove_item(0)
        self.ui.dl_live_wdg.clear()
        self.ui.actionCa_ncel.setEnabled(False)
        if not fname:
            self.handle_downloading_error()
            return

        try:
            os.rename(fname, self.video_loaded)
            lgg.info('File renamed: %s' % os.path.basename(self.video_loaded))
        except Exception as why:
            lgg.warning("Can't rename %s" % self.video_loaded)
            lgg.warning("Reason: %s" % why)

        self.check_file_size()
        self.copy_summary(item, self.video_loaded)
        self.copy_thumbnail(item, self.video_loaded)
        if self.to_download:
            self.start_download()

    def check_file_size(self):
        if self.loading_aborted:
            # The downloading was stopped by the user
            self.loading_aborted = False
            return

        expected = self.loader.size
        loaded = os.path.getsize(self.video_loaded)
        if loaded != expected:
            # TODO format numbers
            lgg.warning("The downloading seems incomplete!")
            lgg.warning("Size expected: %s downloaded: %s" %(expected, loaded))
            fname = os.path.basename(self.video_loaded)
            warn = WarningBox(9, (fname, expected, loaded), self.ui)
            rep = warn.exec_()

    def copy_summary(self, video, fname):
        """Save the summary into a file.

        Args:
        video -- TVItem instance
        fname -- path-name of the video downloaded
        """
        if not self.core.cfg['pitch_live']:
            return

        if self.core.cfg['pitch_live_unique']:
            path = os.path.join(os.path.dirname(fname), 'index')

        else:
            path = os.path.splitext(fname)[0]

        title = video.fname
        summary = self.get_description(video)
        txt = "\n* %s\n  %s\n" %(title, summary)
        try:
            with open(path, 'a') as outf:
                outf.write(txt)
                lgg.info("Summary saved")
        except Exception as why:
            lgg.warning("Can't write summary: %s" % path)
            lgg.warning("Reason: %s" % why)

    def copy_thumbnail(self, video, fname):
        """Save the thumbnail into a file.

        Args:
        video -- TVItem instance
        fname -- path-name of the video downloaded
        """
        if self.core.cfg['copythumb_plus']:
            path = os.path.splitext(fname)[0] + ".jpg"
            orig = video.pixmap
            shutil.copy(orig, path)
            lgg.info("Thumbnail saved")

    def handle_downloading_error(self):
        """Display a warning box when the downloading failed.

        """
        url = self.loader.url
        error = self.loader.last_error
        lgg.warning("Downloading error: %s" % error)
        lgg.warning("URL: %s" % url)
        warn = WarningBox(8, (url, error), self.ui)
        rep = warn.exec_()
        if self.ui.tv_basket.lst_movies:
            self.start_download()

    def get_description(self, item):
        if not item.summary:
            return item.teaser

        return item.summary

    def download_later(self, index):
        lgg.info('Call deferred task with item %s' % index)
        video = self.viewer.videos_items[index]
        qualities = self.get_stream_urls(video)

        if qualities is None:
            lgg.warning("No stream URLs found, item rejected")
            return

        self.clean_qualities(qualities)
        video.qualities = qualities
        video.set_file_name()
        video.set_quality(self.default_quality)
        self.core.create_deferred_task(video)
        

class ConcertItem(object):
    def __init__(self, cat, item, lang):
        """Item concert in arte Live Web.

        Args:
        cat -- category
        item -- values provided by the parser
        lang -- user langage ('fr' or 'de')
        """
        self.category = cat
        self.order = item['id']
        self.title = item['title']
        self.date = item['date']
        self.expires = item['expires']
        self.duration = item['duration']
        self.teaser = item['teaser']
        self.url = item['url']
        self.imgurl = item['imgurl']
        self.jsonurl = item['jsonurl']
        self.preview = None
        self.qualities = None
        self.quality = None
        self.lang = lang
        self.fname = None
        self.author = None
        try:
            self.summary = item['summary']
        except:
            self.summary = ""

    def set_file_name(self):
        f = self.title.replace('/', '-').replace('»', '_').replace('«', '_')
        f = f.replace('\\', ' ').replace('.', ' ')
        if len(f) > 48:
            f = "%s---%s" %(f[:35], f[-10:])

        self.fname = f

    def set_quality(self, quality):
        if not quality in self.qualities:
            for q in ['HTTP_EQ_1', 'HTTP_HQ_1', 'HTTP_SQ_1', 'HTTP_MQ_1']:
                if q in self.qualities:
                    quality = q
                    break

        self.quality = quality

    def get_image(self):
        if self.preview is not None:
            try:
                img = urllib2.urlopen(self.pix).read()
            except Exception as why:
                return None

            else:
                return img

        else:
            return None

    def format_date(self):
        return self.date.replace('T', ' ').rstrip('Z')

    def format_duration(self):
        min_, sec = divmod(int(self.duration), 60)
        if min_ < 60:
            return "%s min. %s sec." %(min_, sec)

        hours, min_ = divmod(min_, 60)
        return "%s h. %s min. %s sec." %(hours, min_, sec)


class Parser(HTMLParser):
    """Define the parser of arte Concert's pages.

    """
    def __init__(self, driver, urls, last_index):
        super().__init__()
        self.driver = driver
        self.site_pages = urls
        self.last_index = last_index
        self.inarticle = False
        self.concert = False
        self.indescription = False
        self.all_concerts = []
        self.concerts = []
        self.indexes = []
        self.stop = False

    def handle_starttag(self, tag, attrs):
        if tag == "article":
            self.inarticle = True
            for att in attrs:
                if att[0] == 'class':
                    try:
                        data = att[1].replace('\n', ' ')
                        idx = data.split('nid-')[1].split(' ')[0].strip()
                        if idx not in self.indexes:
                            self.indexes.append(idx)
                            self.concert = {'id': idx}
                    except IndexError:
                        pass
                    
        elif self.inarticle and self.concert:
            if tag == "a":
                for att in attrs:
                    if att[0] == "title":
                        self.concert['title'] = att[1]
                    elif att[0] == "href":
                        try:
                            ref = att[1].split('view-source:')[1].split(">")
                            self.concert['url'] = ref
                        except IndexError:
                            ref = att[1]
                            self.concert['url'] = ref
                            
                            #print('Href: %s' % ref)
                            # get ref[0] page 
                            # find <div class="video-container"
                            #arte_vp_config="arte_concert"
                            #arte_vp_lang="fr_FR"
                            #arte_vp_url="http://concert.arte.tv/fr/player/55886"
                            # get page http://concert.arte.tv/fr/player/xxxxx
                            # find the json "VSR":

            elif tag == 'div':
                try:
                    if attrs[1] == ('property', 'content:encoded'):
                        self.indescription = True
                except IndexError:
                    pass

    def handle_endtag(self, tag):
        if tag == "article":
            self.inarticle = False
            self.concert = False

    def handle_data(self, data):
        if self.indescription:
            self.concert['teaser'] = data
            self.concerts.append(self.concert)
            self.indescription = False
            self.concert = False

    def parse_all_pages(self):
        for page in self.site_pages:
            self.concerts = []
            if self.stop:
                break

            content = self.fetch_page(page)
            if content:
                self.feed(content)
                
            self.all_concerts.append(self.concerts)

        self.driver.on_parsing_finished()

    def fetch_page(self, url):
        """Fetch a html or xml page.

        Args:
        url -- page URL

        Returns:
        page as text or False if fail
        """
        lgg.info('Load page: %s' % url)
        try:
            content = urllib.request.urlopen(url).read()
            return str(content.decode('utf-8', 'replace'))
        except Exception as why:
            lgg.info('urllib2 error: %s, %s' % (url, why))
            return False

    def get_new_videos(self, items):
        for idx, item in enumerate(items):
            if item['id'] == self.last_index:
                self.stop = True
                break

        return items


class Core:
    def __init__(self):
        self.ui = None
        self.cfg = {'lang': 'fr'}
        self.workspace = {'live_tmp': '/home/vincent/qarte-3/wspace/livePreviews'}

        self.arteconcert = ArteConcert(self)
        self.arteconcert.config_parser()

if __name__ == '__main__':
    core = Core()
