# -*- encoding: utf-8 -*-

# artelivewebviewer.py
# This file is part of Qarte-3
#    
# Author: Vincent Vande Vyvre <vincent.vandevyvre@oqapy.eu>
# Copyright: 2011-2016 Vincent Vande Vyvre
# Licence: GPL3
# Home page: https://launchpad.net/qarte
#
# Graphics view for 'arte Live Web' page

import os
import gettext
import logging
lgg = logging.getLogger(__name__)

from PyQt5 import QtCore, QtGui, QtWidgets


#from themes import Themes


class ArteLiveWebViewer(QtWidgets.QGraphicsView):
    def __init__(self, ui, parent=None):
        super(ArteLiveWebViewer, self).__init__(parent)
        self.setMinimumSize(QtCore.QSize(0, 250))
        self.setMaximumSize(QtCore.QSize(16777215, 270))
        self.setDragMode(QtWidgets.QGraphicsView.ScrollHandDrag)
        self.setTransformationAnchor(QtWidgets.QGraphicsView.AnchorViewCenter)
        self.setStyleSheet("""QGraphicsView {color: white;
                            background: qlineargradient(
                            x1: 0, y1: 0, x2: 0, y2: 1,
                            stop: 0 #FFFFFF,
                            stop: 0.4 #606060, stop: 0.5 #101010,
                            stop: 1 #000000)}""")
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)

        self.ui = ui
        self.al = None
        self.prev_fld = None
        self.cfg = None
        self.videos_items = None
        self.news = []

    def show_category(self, videos):
        """Shows preview items in the viewer area.

        """
        self.images = []
        self.videos_items = videos
        for idx, vid in enumerate(self.videos_items):
            is_preview = vid.preview
            if not os.path.isfile(vid.preview):
                img = os.path.join("medias", "noPreview.png")

            else:
                img = vid.preview

            self.images.append(img)

        self.prepare_scene()
        self.populate_scene()

    def prepare_scene(self):
        """Create the QGraphicsScene.

        """
        mg = 20 # margin
        sp = 10 # spacing
        nb = len(self.images)
        lng = mg + ((nb - 1) * (self.cfg['mini'] + 10)) + self.cfg['maxi'] + mg
        # TODO Scene height must be in config

        try:
            self.scene.clear()
        except:
            self.scene = QtWidgets.QGraphicsScene()

        self.scene.setSceneRect(0.0, 0.0, float(lng), self.height()-5)
        self.setScene(self.scene)

    def populate_scene(self):
        """Create and set thumbnails into the scene.

        """
        if not self.videos_items:
            lgg.warning("No such videos found")
            return

        posX, posY = 20, 0
        self.previews = []
        self.current = 0
        item = PreviewItem(self, self.images[0], 0,
                            self.videos_items[0].title, 
                            self.videos_items[0].teaser,
                            self.cfg['maxi'], None, self.scene)
        y = (self.scene.height() - item.height_) / 2
        item.setPos(posX, y)
        self.al.show_pitch(self.videos_items[0])
        QtCore.QCoreApplication.processEvents()
        self.previews.append(item)
        posX += self.cfg['maxi'] + 10
        posY += 40

        for idx, img in enumerate(self.images):
            if idx:
                item = PreviewItem(self, self.images[idx], 
                                    idx,
                                    self.videos_items[idx].title, 
                                    self.videos_items[idx].teaser,
                                    self.cfg['mini'], None, self.scene)
                item.setPos(posX, posY)
                QtCore.QCoreApplication.processEvents()
                self.previews.append(item)
                posX += self.cfg['mini'] + 10
                QtCore.QCoreApplication.processEvents()

        self.center_item()

    def shift_to_next(self):
        """Move one step to the right.

        """
        nb = len(self.images) - 1
        self.minimise_current()
        self.current = self.current + (1 if self.current < nb else -nb)
        self.maximise_current()

        if self.current:
            self.previews[self.current].moveBy(self.cfg['mini'] - 
                                                self.cfg['maxi'], 0)

        else:
            # If we're come back to the first image, we need to move
            # all others pixmap to the right
            for p in self.previews[1:]:
                p.moveBy(self.cfg['maxi'] - self.cfg['mini'], 0)

        self.center_item()

    def shift_to_previous(self):
        """Move one step to the left.

        """
        nb = len(self.images) - 1
        self.minimise_current()
        self.previews[self.current].moveBy(self.cfg['maxi'] - 
                                            self.cfg['mini'], 0)

        if not self.current:
            for p in self.previews:
                p.moveBy(self.cfg['mini'] - self.cfg['maxi'], 0)
            
        self.current = self.current - (1 if self.current > 0 else -nb)
        self.maximise_current()
        self.center_item()

    def go_to_item(self, idx):
        """Move to item idx.

        Called when an item has been clicked.

        Args:
        idx -- index of item
        """
        if idx == self.current:
            return

        delta = self.cfg['maxi'] - self.cfg['mini']
        self.minimise_current()

        if idx > self.current:
            for i in range(self.current, idx):
                self.previews[i + 1].moveBy(-delta, 0)

        else:
            for i in range(idx, self.current):
                self.previews[i + 1].moveBy(delta, 0)

        self.current = idx
        self.maximise_current()
        self.center_item()

    def on_item_selected(self, idx):
        """Check if selection mode is interactiv with deferred download
        dialog.

        Args:
        idx -- item's index
        """
        self.al.core.on_tv_item_selected(self.videos_items[idx])

    def minimise_current(self):
        """Reduce the size of a preview item.

        """
        pos = self.previews[self.current].pos()
        teaser = self.previews[self.current].teaser
        self.scene.removeItem(self.previews.pop(self.current))
        item = PreviewItem(self, self.images[self.current],
                            self.current,
                            self.videos_items[self.current].title, 
                            teaser, self.cfg['mini'], None, self.scene)
        item.setPos(pos.x(), 40)
        self.previews.insert(self.current, item)

    def maximise_current(self):
        """Enlarge the size of a preview item.

        """
        pos = self.previews[self.current].pos()
        teaser = self.previews[self.current].teaser
        self.scene.removeItem(self.previews.pop(self.current))
        item = PreviewItem(self, self.images[self.current],
                            self.current,
                            self.videos_items[self.current].title,
                            teaser, self.cfg['maxi'],  None, self.scene)
        y = (self.scene.height() - item.height_) / 2
        item.setPos(pos.x(), y)
        self.previews.insert(self.current, item)
        self.al.show_pitch(self.videos_items[self.current])

    def center_item(self):
        """Move the current preview item at the center of the viewer area.

        """
        self.centerOn(self.previews[self.current])
        self.al.on_thumbnail_selected(self.videos_items[self.current])
        # Check if video is already in the downloading basket
        self.ui.select_btn.setEnabled(not self.videos_items[self.current]
                                        in self.al.to_download)
        # and if already available
        #self.ui.select_btn.setEnabled(not self.videos_items[self.current].soon)


class PreviewItem(QtWidgets.QGraphicsPixmapItem):
    def __init__(self, main, img, idx, title, teaser, width, 
                    parent=None, scene=None):
        super(PreviewItem, self).__init__(parent)
        if scene is not None: 
            scene.addItem(self)
        self.main = main
        self.img = img
        self.idx = idx
        self.title = title
        self.teaser = teaser
        self.width_ = width
        self.height_ = 0
        self.pix = None
        self.font = QtGui.QFont()
        self.font.setFamily("Verdana")
        self.font.setPointSize(10)
        self.font.setItalic(True)
        self.font_1 = QtGui.QFont()
        self.font_1.setFamily("Verdana")
        self.font_1.setPointSize(20)
        self.font_1.setWeight(55)
        self.font_1.setItalic(True)
        self.is_double_clicked = False
        self.format_title()
        self.__set_image()
        self.__set_frame()
        self.setPixmap(self.pix)
        self.setToolTip(teaser)

    def __set_image(self):
        self.base = QtGui.QImage(self.img).scaledToWidth(self.width_, 
                                QtCore.Qt.SmoothTransformation)

        if self.base.isNull():
            self.base = QtGui.QImage("medias/noPreview.png").scaledToWidth(
                                     self.width_, 
                                     QtCore.Qt.SmoothTransformation)

        if self.title is None:
            self.height_ = self.base.height()

        else:
            self.height_ = self.base.height() + (20 * len(self.title))

    def __set_frame(self):
        self.pix = self.__set_pix()
        rect = QtCore.QRectF(0.0, 0.0, self.width_, self.base.height())
        painter = QtGui.QPainter()
        painter.begin(self.pix)
        painter.setRenderHints(QtGui.QPainter.Antialiasing, True)
        self.path = QtGui.QPainterPath()
        self.path.addRoundedRect(rect, 6.0, 6.0)
        painter.drawPath(self.path)
        painter.end()
        self.__fill_path()

    def __fill_path(self):
        """Draw the texture into the shape.

        """
        painter = QtGui.QPainter()
        painter.begin(self.pix)
        painter.setRenderHints(QtGui.QPainter.Antialiasing, True)
        brush = QtGui.QBrush()
        brush.setTexture(self.__set_texture())     
        painter.fillPath(self.path, brush)
        painter.end()

        if self.title is not None:
            self.write_title()

    def __set_texture(self):
        texture = QtGui.QPixmap(QtCore.QSize(self.width_, self.base.height()))
        irect = QtCore.QRect(0, 0, self.width_, self.base.height())
        prect = QtCore.QRect(0, 0, self.width_, self.base.height())
        painter = QtGui.QPainter()
        painter.begin(texture)
        painter.drawImage(prect, self.base, irect)
        painter.end()

        return texture

    def __set_pix(self):
        """Create the pixmap support for the texture.

        Returns:
        QPixmap square filled transparent
        """
        color = QtGui.QColor()
        color.setRgb(0, 0, 0, 0)
        pix = QtGui.QPixmap(QtCore.QSize(self.width_, self.height_))
        pix.fill(color)

        return pix

    def format_title(self):
        if self.title is None:
            self.title = "No Title"
        title = Title(self)
        lines = title.partition_text()

        if lines is None:
            self.title = ["No Title"]

        else:
            self.title = lines

    def write_title(self):
        y = self.base.height() + 15
        painter = QtGui.QPainter()
        painter.begin(self.pix)
        painter.setPen(QtGui.QPen(QtCore.Qt.white, 2,
                                  QtCore.Qt.SolidLine, QtCore.Qt.RoundCap,
                                  QtCore.Qt.MiterJoin))
        painter.setFont(self.font)

        for i in self.title:
            l = self.get_text_length(i)
            x = (self.width_ / 2) - (l / 2)
            painter.drawText(x, y, i)
            y += 15

        painter.end()

    def __set_label_new(self):
        if self.new == -1:
            label = "Soon"
            x = self.width_ - 90

        else:
            label = "New"
            x = self.width_ - 70

        y = self.base.height()
        painter = QtGui.QPainter()
        painter.begin(self.pix)
        painter.setPen(QtGui.QPen(QtCore.Qt.red, 2,
                                  QtCore.Qt.SolidLine, QtCore.Qt.RoundCap,
                                  QtCore.Qt.MiterJoin))
        painter.setFont(self.font_1)
        painter.drawText(x, y, label)
        painter.end()

    def get_text_length(self, t):
        """Get string's lenght with the help of font metrics.

        Keyword arguments:
        t -- text
        
        Returns:
        lenght of text
        """
        metric = QtGui.QFontMetrics(self.font)
        return metric.width(t)

    def mousePressEvent(self, event):
        if event.button() == 1:
            timer = QtCore.QTimer()
            timer.singleShot(300, self.on_item_clicked)
            timer.start()

    def mouseDoubleClickEvent(self, event):
        self.is_double_clicked = True
        self.main.al.select_video_to_download(self.idx)

    def on_item_clicked(self):
        # The item may be destroyed and replaced with a large one
        # In case of double click the mousePressEvent is emitted twice,
        # may be a bug in Qt5 ...
        if self is not None:
            if not self.is_double_clicked:
                self.main.go_to_item(self.idx)
                self.main.on_item_selected(self.idx)

            self.is_double_clicked = False

    def contextMenuEvent(self, event):
        self.main.go_to_item(self.idx)
        self.show_menu(event.screenPos())
        event.accept()

    def show_menu(self, pos):
        """Build and show context menu.

        Keyword arguments:
        pos -- event.pos()
        """
        self.menu = QtWidgets.QMenu()
        self.act_add = QtWidgets.QAction(self.menu)
        self.act_add.setText(_("Add to download list"))
        self.menu.addAction(self.act_add)
        self.act_diff = QtWidgets.QAction(self.menu)
        self.act_diff.setText(_("Deferred download"))
        self.menu.addAction(self.act_diff)
        self.act_add.triggered.connect(self.add_to_download)
        self.act_diff.triggered.connect(self.set_cron_job)

        #if self.new == -1:
            #self.act_add.setEnabled(False)

        #self.menu.setStyleSheet(self.main.ui.menu_style)
        self.menu.popup(pos)

    def add_to_download(self):
        QtCore.QCoreApplication.processEvents()
        self.main.al.select_video_to_download(self.idx)

    def set_cron_job(self):
        QtCore.QCoreApplication.processEvents()
        self.main.al.download_later(self.idx)

class Title(QtWidgets.QGraphicsTextItem):
    def __init__(self, pix, parent=None):
        super(Title, self).__init__(parent)
        self.pix = pix
        self.text_ = pix.title
        self.font = pix.font
        self.width_ = pix.width_ - 20

    def get_text_length(self, t):
        """Get string's lenght with the help of QFontMetrics.

        Keyword arguments:
        t -- text
        
        Returns:
        lenght of text
        """
        metric = QtGui.QFontMetrics(self.font)
        return metric.width(t)

    def partition_text(self):
        """Divide the title in a list of lines.

        Returns:
        list of lines
        """
        def adjust_length(ch):
            while self.get_text_length(ch) >= self.width_:
                ch = ch[:-1]

            return len(ch)

        mtc = self.get_text_length(self.text_)
        #print(self.text_)
        #print('Orig lenght: %s for 53 chars' % mtc)
        if mtc <= self.width_:
            return [self.text_]

        parts = []
        begin = 0
        lc = int(len(self.text_) / (mtc / self.width_))
        chain = self.text_[:lc] 
        #print('First iter with %s len: %s' % (chain, len(chain)))
        while 1:
            end = adjust_length(chain)
            #print('%s adusted: %s, %s' % (self.text_, end, len(self.text_)))
            if len(self.text_) == end:
                part = chain[:end]

            elif not self.text_[end] in [" ", "-", "_"]:
                part = self.find_caesura(chain[:end])
                #print('return %s' % part)

            else:
                part = chain[:end]

            parts.append(part)
            self.text_ = self.text_.replace(part, "").lstrip()

            if len(self.text_) >= lc:
                chain = self.text_[:lc]

            else:
                parts.append(self.text_)
                break

        return parts

    def find_caesura(self, chain):
        """Find the appropriate caesura of a string.

        Args:
        chain -- part of video's title (one line)

        Returns:
        Formated text
        """
        count = len(chain) - 1
        while count > 0:
            if chain[count] in [" ", "-", "_"]:
                break
            count -= 1

        if count:
            return chain[:count]

        return chain

